#!/usr/bin/env python
#
# -----------------------------------------------------------------------------
# Copyright (C) 2016 Daniel Standage <daniel.standage@gmail.com>
#
# This file is part of tag (http://github.com/standage/tag) and is licensed
# under the BSD 3-clause license: see LICENSE.
# -----------------------------------------------------------------------------

from __future__ import print_function
import tag


mains = {
    'gff3': tag.cli.gff3.main,
    'occ': tag.cli.occ.main,
}


def main(args=None):
    """
    Entry point for the tag CLI.

    Isolated as a method so that the CLI can be called by other Python code
    (e.g. for testing), in which case the arguments are passed to the function.
    If no arguments are passed to the function, parse them from the command
    line.
    """
    if args is None:
        args = tag.cli.parser().parse_args()

    assert args.cmd in mains
    mainmethod = mains[args.cmd]
    mainmethod(args)



#!/usr/bin/env python
#
# -----------------------------------------------------------------------------
# Copyright (C) 2016 Daniel Standage <daniel.standage@gmail.com>
#
# This file is part of tag (http://github.com/standage/tag) and is licensed
# under the BSD 3-clause license: see LICENSE.
# -----------------------------------------------------------------------------

import argparse
import sys
import tag
from . import gff3
from . import occ


def parser():
    parser = argparse.ArgumentParser()
    parser.add_argument('-v', '--version', action='version',
                        version='tag v{}'.format(tag.__version__))
    parser.add_argument('-l', '--logfile', metavar='FILE', default=sys.stderr,
                        type=argparse.FileType('w'))
    subparsers = parser.add_subparsers(dest='cmd', metavar='cmd',
                                       help='gff3 | occ')
    tag.cli.gff3.subparser(subparsers)
    tag.cli.occ.subparser(subparsers)

    return parser



#!/usr/bin/env python
#
# -----------------------------------------------------------------------------
# Copyright (C) 2016 Daniel Standage <daniel.standage@gmail.com>
#
# This file is part of tag (http://github.com/standage/tag) and is licensed
# under the BSD 3-clause license: see LICENSE.
# -----------------------------------------------------------------------------

import tag


def features(entrystream, type=None, traverse=False):
    """
    Pull features out of the specified entry stream.

    :param entrystream: a stream of entries
    :param type: retrieve only features of the specified type; set to
                 :code:`None` to retrieve all features
    :param traverse: by default, only top-level features are selected; set
                     to :code:`True` to search each feature graph for the
                     specified feature type
    """
    for feature in entry_type_filter(entrystream, tag.feature.Feature):
        if traverse:
            if type is None:
                message = 'cannot traverse without a specific feature type'
                raise ValueError(message)
            if type == feature.type:
                yield feature
            else:
                for subfeature in feature:
                    if type == subfeature.type:
                        yield subfeature
        else:
            if not type or type == feature.type:
                yield feature


def window(featurestream, seqid, start=None, end=None, strict=True):
    """
    Pull features out of the designated genomic interval.

    This function uses 0-based half-open intervals, not the 1-based closed
    intervals used by GFF3.

    :param featurestream: a stream of feature entries
    :param seqid: ID of the sequence from which to select features
    :param start: start of the genomic interval
    :param end: end of the genomic interval
    :param strict: when set to :code:`True`, only features completely contained
                   within the interval are selected; when set to :code:`False`,
                   any feature overlapping the interval is selected
    """
    region = None
    if start and end:
        region = tag.range.Range(start, end)

    for feature in featurestream:
        if feature.seqid != seqid:
            continue
        if region:
            if strict:
                if region.contains(feature._range):
                    yield feature
            else:
                if region.overlap(feature._range):
                    yield feature
        else:
            yield feature


def directives(entrystream, type=None):
    """
    Pull directives out of the specified entry stream.

    :param entrystream: a stream of entries
    :param type: retrieve only directives of the specified type; set to
                 :code:`None` to retrieve all directives
    """
    for directive in entry_type_filter(entrystream, tag.directive.Directive):
        if not type or type == directive.type:
            yield directive


def sequences(entrystream):
    """Pull sequences out of the specified entry stream."""
    for sequence in entry_type_filter(entrystream, tag.sequence.Sequence):
        yield sequence


def entry_type_filter(entrystream, entryclass):
    """
    Generic entry filter.

    :param entrystream: a stream of entries
    :param entryclass: specify the type of entry upon which to filter
    """
    for entry in entrystream:
        if isinstance(entry, entryclass):
            yield entry



#!/usr/bin/env python
#
# -----------------------------------------------------------------------------
# Copyright (C) 2015 Daniel Standage <daniel.standage@gmail.com>
#
# This file is part of tag (http://github.com/standage/tag) and is licensed
# under the BSD 3-clause license: see LICENSE.
# -----------------------------------------------------------------------------

import argparse
import tag


def subparser(subparsers):
    subparser = subparsers.add_parser('gff3')
    subparser.add_argument('-o', '--out', metavar='FILE',
                           default='/dev/stdout', help='write output to the '
                           'specified file; default is terminal (stdout)')
    subparser.add_argument('-r', '--relax', action='store_false', default=True,
                           dest='strict', help='relax parsing stringency')
    subparser.add_argument('gff3', help='input file in GFF3 format')


def main(args):
    reader = tag.reader.GFF3Reader(infilename=args.gff3, strict=args.strict)
    writer = tag.writer.GFF3Writer(reader, args.out)
    writer.write()



#!/usr/bin/env python
#
# -----------------------------------------------------------------------------
# Copyright (C) 2015 Daniel Standage <daniel.standage@gmail.com>
#
# This file is part of tag (http://github.com/standage/tag) and is licensed
# under the BSD 3-clause license: see LICENSE.
# -----------------------------------------------------------------------------

from __future__ import print_function
import argparse
from collections import defaultdict
from intervaltree import IntervalTree
import tag


def interval_set_span(intset):
    begin = min([x for x, y, z in intset])
    end = max([y for x, y, z in intset])
    feats = set([z for x, y, z in intset])
    return begin, end, feats


def subparser(subparsers):
    subparser = subparsers.add_parser('occ')
    subparser.add_argument('-r', '--relax', action='store_false', default=True,
                           dest='strict', help='relax parsing stringency')
    subparser.add_argument('gff3', help='input file')
    subparser.add_argument('type', help='feature type')


def main(args):
    features = defaultdict(IntervalTree)
    reader = tag.reader.GFF3Reader(infilename=args.gff3, strict=args.strict)
    for feature in tag.select.features(reader, type=args.type, traverse=True):
        features[feature.seqid].addi(feature.start, feature.end, feature)
        if feature.is_multi and feature.is_toplevel:
            for sib in feature.siblings:
                features[sib.seqid].addi(sib.start, sib.end, sib)

    total_occ = 0
    ints_acct_for = defaultdict(IntervalTree)
    for seqid in features:
        for interval in features[seqid]:
            begin, end, feat = interval
            if ints_acct_for[seqid][begin:end] != set():
                continue

            feats = set([feat])
            overlapping = features[seqid][begin:end]
            testbegin, testend, testfeats = interval_set_span(overlapping)
            while set(feats) < testfeats:
                begin, end, feats = testbegin, testend, testfeats
                overlapping = features[seqid][begin:end]
                testbegin, testend, testfeats = interval_set_span(overlapping)
            total_occ += end - begin
            ints_acct_for[seqid].addi(begin, end, feats)
    print(total_occ)
